/*
 *        nProbe - a Netflow v5/v9/IPFIX probe for IPv4/v6
 *
 *       Copyright (C) 2002-12 Luca Deri <deri@ntop.org>
 *
 *                     http://www.ntop.org/
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License along
 *  with this program; if not, write to the Free Software Foundation, Inc.,
 *  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 */

#include "nprobe.h"

/* ************************************ */

static void incrementQueueStats(void) {
#ifdef HAVE_REDIS
  readOnlyGlobals.redis.queuedSetDeleteCommands++;
  
  if(readOnlyGlobals.redis.queuedSetDeleteCommands > readOnlyGlobals.redis.maxQueuedSetDeleteCommands)
    readOnlyGlobals.redis.maxQueuedSetDeleteCommands = readOnlyGlobals.redis.queuedSetDeleteCommands;
#endif
}

/* ************************************ */

void setCacheKeyValueString(const char *prefix, const char *key, const char *value) {
#ifdef HAVE_REDIS
  if(readOnlyGlobals.redis.context) {
    pthread_rwlock_wrlock(&readOnlyGlobals.redis.lock_set_delete);
    if(unlikely(readOnlyGlobals.enable_debug)) traceEvent(TRACE_NORMAL, "[Redis] SET %s%s %s", prefix, key, value);
    redisAppendCommand(readOnlyGlobals.redis.set_delete_context, "SET %s%s %s", prefix, key, value);
    incrementQueueStats();
    pthread_rwlock_unlock(&readOnlyGlobals.redis.lock_set_delete);
  }
#endif
}

/* ************************************************* */

void setCacheKeyValueNumber(const char *prefix, const char *key, u_int64_t value) {
#ifdef HAVE_REDIS
  if(readOnlyGlobals.redis.context) {
    pthread_rwlock_wrlock(&readOnlyGlobals.redis.lock_set_delete);
    if(unlikely(readOnlyGlobals.enable_debug)) traceEvent(TRACE_NORMAL, "[Redis] SET %s%s %llu", prefix, key, value);
    redisAppendCommand(readOnlyGlobals.redis.set_delete_context, "SET %s%s %llu", prefix, key, value);
    incrementQueueStats();
    pthread_rwlock_unlock(&readOnlyGlobals.redis.lock_set_delete);
  }
#endif
}

/* ************************************************* */

void incrCacheKeyValueNumber(const char *prefix, const char *key, u_int64_t value) {
#ifdef HAVE_REDIS
  if(readOnlyGlobals.redis.context) {
    pthread_rwlock_wrlock(&readOnlyGlobals.redis.lock_set_delete);
    if(unlikely(readOnlyGlobals.enable_debug)) traceEvent(TRACE_NORMAL, "[Redis] INCRBY %s%s %llu", prefix, key, value);
    redisAppendCommand(readOnlyGlobals.redis.set_delete_context, "INCRBY %s%s %llu", prefix, key, value);
    incrementQueueStats();
    pthread_rwlock_unlock(&readOnlyGlobals.redis.lock_set_delete);
  }
#endif
}

/* ************************************************* */

void incrHashCacheKeyValueNumber(const char *element, const char *key, u_int64_t value) {
#ifdef HAVE_REDIS
  if(readOnlyGlobals.redis.context) {
    pthread_rwlock_wrlock(&readOnlyGlobals.redis.lock_set_delete);
    if(unlikely(readOnlyGlobals.enable_debug)) traceEvent(TRACE_NORMAL, "[Redis] HINCRBY %s %s %llu", element, key, value);
    redisAppendCommand(readOnlyGlobals.redis.set_delete_context, "HINCRBY %s %s %llu", element, key, value);
    incrementQueueStats();
    pthread_rwlock_unlock(&readOnlyGlobals.redis.lock_set_delete);
  }
#endif
}

/* ************************************************* */

void expireCacheKey(const char *prefix, const char *key, u_int32_t duration_sec) {
#ifdef HAVE_REDIS
  if(readOnlyGlobals.redis.context) {
    pthread_rwlock_wrlock(&readOnlyGlobals.redis.lock_set_delete);
    if(unlikely(readOnlyGlobals.enable_debug)) traceEvent(TRACE_NORMAL, "[Redis] EXPIRE %s%s %u", prefix, key, duration_sec);
    redisAppendCommand(readOnlyGlobals.redis.set_delete_context, "EXPIRE %s%s %u", prefix, key, duration_sec);
    incrementQueueStats();
    pthread_rwlock_unlock(&readOnlyGlobals.redis.lock_set_delete);
  }
#endif
}

/* ************************************ */

void setCacheHashKeyValueString(const char *element, const char *key, const char *value) {
#ifdef HAVE_REDIS
  if(readOnlyGlobals.redis.context) {
    pthread_rwlock_wrlock(&readOnlyGlobals.redis.lock_set_delete);
    if(unlikely(readOnlyGlobals.enable_debug)) traceEvent(TRACE_NORMAL, "[Redis] HSET %s %s %s", element, key, value);
    redisAppendCommand(readOnlyGlobals.redis.set_delete_context, "HSET %s %s %s", element, key, value);
    incrementQueueStats();
    pthread_rwlock_unlock(&readOnlyGlobals.redis.lock_set_delete);
  }
#endif
}

/* ************************************************* */

void setCacheHashKeyValueNumber(const char *element, const char *key, u_int64_t value) {
#ifdef HAVE_REDIS
  if(readOnlyGlobals.redis.context) {
    pthread_rwlock_wrlock(&readOnlyGlobals.redis.lock_set_delete);
    if(unlikely(readOnlyGlobals.enable_debug)) traceEvent(TRACE_NORMAL, "[Redis] HSET %s %s %llu", element, key, value);
    redisAppendCommand(readOnlyGlobals.redis.set_delete_context, "HSET %s %s %llu", element, key, value);
    incrementQueueStats();
    pthread_rwlock_unlock(&readOnlyGlobals.redis.lock_set_delete);
  }
#endif
}

/* ************************************************* */

void incrCacheHashKeyValueNumber(const char *element,
				 const char *key, 
				 const u_int64_t value) {
#ifdef HAVE_REDIS
  if(readOnlyGlobals.redis.context && (value > 0)) {
    pthread_rwlock_wrlock(&readOnlyGlobals.redis.lock_set_delete);
    if(unlikely(readOnlyGlobals.enable_debug)) traceEvent(TRACE_NORMAL, "[Redis] HINCRBY %s %s %llu", element, key, value);
    redisAppendCommand(readOnlyGlobals.redis.set_delete_context, "HINCRBY %s %s %llu", element, key, value);
    incrementQueueStats();
    pthread_rwlock_unlock(&readOnlyGlobals.redis.lock_set_delete);
  }
#endif
}

/* ************************************************* */

void zIncrCacheHashKeyValueNumber(const char *set_name,
				  const char *key, 
				  const u_int64_t value) {
#ifdef HAVE_REDIS
  if(readOnlyGlobals.redis.context && (value > 0)) {
    pthread_rwlock_wrlock(&readOnlyGlobals.redis.lock_set_delete);
    if(unlikely(readOnlyGlobals.enable_debug)) traceEvent(TRACE_NORMAL, "[Redis] ZINCRBY %s %llu %s", set_name, value, key);
    redisAppendCommand(readOnlyGlobals.redis.set_delete_context, "ZINCRBY %s %llu %s", set_name, value, key);
    readOnlyGlobals.redis.queuedSetDeleteCommands++;
    pthread_rwlock_unlock(&readOnlyGlobals.redis.lock_set_delete);
  }
#endif
}

/* ************************************************* */

void setCacheNumKeyMixedValueQuad(const char *prefix,
                                  const u_int32_t key0, const char* value0,
                                  const u_int32_t key1, const char* value1,
                                  const u_int32_t key2, const u_int32_t value2,
                                  const u_int32_t key3, const u_int32_t value3) {
#ifdef HAVE_REDIS
  if(readOnlyGlobals.redis.context) {
    pthread_rwlock_wrlock(&readOnlyGlobals.redis.lock_set_delete);
    if(unlikely(readOnlyGlobals.enable_debug))
      traceEvent(TRACE_NORMAL, "[Redis] MSET %s%u %s %s%u %s %s%u %u %s%u %u",
                 prefix, key0, value0, prefix, key1, value1,
                 prefix, key2, value2, prefix, key3, value3);
    redisAppendCommand(readOnlyGlobals.redis.set_delete_context,
                       "MSET %s%u %s %s%u %s %s%u %u %s%u %u",
                       prefix, key0, value0, prefix, key1, value1,
                       prefix, key2, value2, prefix, key3, value3);
    incrementQueueStats();
    pthread_rwlock_unlock(&readOnlyGlobals.redis.lock_set_delete);
  }
#endif
}

/* ************************************************* */

void setCacheHashNumKeyMixedValueQuad(const char *master_key,
				      const u_int32_t key0, const char* value0,
				      const u_int32_t key1, const char* value1,
				      const u_int32_t key2, const u_int32_t value2,
				      const u_int32_t key3, const u_int32_t value3) {
#ifdef HAVE_REDIS
  if(readOnlyGlobals.redis.context) {
    pthread_rwlock_wrlock(&readOnlyGlobals.redis.lock_set_delete);
    if(unlikely(readOnlyGlobals.enable_debug))
      traceEvent(TRACE_NORMAL, "[Redis] HMSET %s %u %s %u %s %u %u %u %u",
		 master_key,
		 key0, value0,
		 key1, value1,
		 key2, value2, 
		 key3, value3);
    redisAppendCommand(readOnlyGlobals.redis.set_delete_context,
		       "HMSET %s %u %s %u %s %u %u %u %u",
		       master_key,
		       key0, value0,
		       key1, value1,
		       key2, value2,
		       key3, value3);

    incrementQueueStats();
    pthread_rwlock_unlock(&readOnlyGlobals.redis.lock_set_delete);
  }
#endif
}

/* ************************************ */

char* getCacheDataNumKey(const char *prefix, const u_int32_t key) {
#ifdef HAVE_REDIS
  char *ret = NULL;
  redisReply *reply;

  if(readOnlyGlobals.redis.context) {
    pthread_rwlock_wrlock(&readOnlyGlobals.redis.lock_get);
    if(unlikely(readOnlyGlobals.enable_debug)) traceEvent(TRACE_NORMAL, "[Redis] GET %s%u", prefix, key);
    reply = redisCommand(readOnlyGlobals.redis.context, "GET %s%u", prefix, key);
    pthread_rwlock_unlock(&readOnlyGlobals.redis.lock_get);

    if(reply) {
      if(reply->str) {
	ret = strdup(reply->str);
	if(unlikely(readOnlyGlobals.enable_debug))
	  traceEvent(TRACE_NORMAL, "[Redis] %s(%u)=%s", __FUNCTION__, key, ret);
      }
      freeReplyObject(reply);
    }
  }

  return(ret);
#else
  return(NULL);
#endif
}

/* ************************************ */

char* getCacheDataStrKey(const char *prefix, const char *key) {
#ifdef HAVE_REDIS
  char *ret = NULL;
  redisReply *reply;

  if(readOnlyGlobals.redis.context) {
    pthread_rwlock_wrlock(&readOnlyGlobals.redis.lock_get);
    if(unlikely(readOnlyGlobals.enable_debug)) traceEvent(TRACE_NORMAL, "[Redis] GET %s%s", prefix, key);
    reply = redisCommand(readOnlyGlobals.redis.context, "GET %s%s", prefix, key);
    pthread_rwlock_unlock(&readOnlyGlobals.redis.lock_get);

    if(reply) {
      if(reply->str) {
	ret = strdup(reply->str);
	traceEvent(TRACE_INFO, "[Redis] %s(%u)=%s", __FUNCTION__, key, ret);
      }
      freeReplyObject(reply);
    }
  }

  return(ret);
#else
  return(NULL);
#endif
}

/* ************************************ */

int deleteCacheStrKey(const char *prefix, const char *key, const u_int32_t delete_delay_sec) {
#ifdef HAVE_REDIS
  if(readOnlyGlobals.redis.context) {
    if(unlikely(readOnlyGlobals.enable_debug)) traceEvent(TRACE_NORMAL, "[Redis] EXPIRE %s%s %d", prefix, key, delete_delay_sec);
    pthread_rwlock_wrlock(&readOnlyGlobals.redis.lock_set_delete);
    if(delete_delay_sec > 0)
      redisAppendCommand(readOnlyGlobals.redis.set_delete_context, "EXPIRE %s%s %d", prefix, key, delete_delay_sec);
    else
      redisAppendCommand(readOnlyGlobals.redis.set_delete_context, "DEL %s%s", prefix, key);

    incrementQueueStats();
    pthread_rwlock_unlock(&readOnlyGlobals.redis.lock_set_delete);
  }
#endif

  return(0);
}

/* ************************************ */

int deleteCacheNumKey(const char *prefix, const u_int32_t key, const u_int32_t delete_delay_sec) {
#ifdef HAVE_REDIS
  if(readOnlyGlobals.redis.context) {
    pthread_rwlock_wrlock(&readOnlyGlobals.redis.lock_set_delete);
    if(unlikely(readOnlyGlobals.enable_debug)) traceEvent(TRACE_NORMAL, "[Redis] EXPIRE %s%u %d", prefix, key, delete_delay_sec);
    if(delete_delay_sec > 0)
      redisAppendCommand(readOnlyGlobals.redis.set_delete_context, "EXPIRE %s%u %d", prefix, key, delete_delay_sec);
    else
      redisAppendCommand(readOnlyGlobals.redis.set_delete_context, "DEL %s%u", prefix, key);

    incrementQueueStats();
    pthread_rwlock_unlock(&readOnlyGlobals.redis.lock_set_delete);
  }
#endif

  return(0);
}

/* ************************************ */

/* Redis does not accept expire time */
int deleteCacheNumKeyTwin(const char *prefix, const u_int32_t key1, const u_int32_t key2) {
#ifdef HAVE_REDIS
  if(readOnlyGlobals.redis.context) {
    pthread_rwlock_wrlock(&readOnlyGlobals.redis.lock_set_delete);
    if(unlikely(readOnlyGlobals.enable_debug)) traceEvent(TRACE_NORMAL, "[Redis] DEL %s%u %s%u", prefix, key1, prefix, key2);
    redisAppendCommand(readOnlyGlobals.redis.set_delete_context, "DEL %s%u %s%u", prefix, key1, prefix, key2);

    incrementQueueStats();
    pthread_rwlock_unlock(&readOnlyGlobals.redis.lock_set_delete);
  }
#endif

  return(0);
}

/* ************************************ */

/* Redis does not accept expire time */
int deleteCacheStrKeyTwin(const char *prefix, const char *key1, const char *key2) {
#ifdef HAVE_REDIS
  if(readOnlyGlobals.redis.context) {
    pthread_rwlock_wrlock(&readOnlyGlobals.redis.lock_set_delete);
    if(unlikely(readOnlyGlobals.enable_debug)) traceEvent(TRACE_NORMAL, "[Redis] DEL %s%s %s%s", prefix, key1, prefix, key2);
    redisAppendCommand(readOnlyGlobals.redis.set_delete_context, "DEL %s%s %s%s", prefix, key1, prefix, key2);

    incrementQueueStats();
    pthread_rwlock_unlock(&readOnlyGlobals.redis.lock_set_delete);
  }
#endif

  return(0);
}

/* ************************************ */

char* getHashCacheDataStrKey(const char *prefix, const char *element, const char *key) {
#ifdef HAVE_REDIS
  char *ret = NULL;
  redisReply *reply;

  if(readOnlyGlobals.redis.context) {
    pthread_rwlock_wrlock(&readOnlyGlobals.redis.lock_get);
    if(unlikely(readOnlyGlobals.enable_debug)) traceEvent(TRACE_NORMAL, "[Redis] HGET %s%s %s", prefix, element, key);
    reply = redisCommand(readOnlyGlobals.redis.context, "HGET %s%s %s", prefix, element, key);
    pthread_rwlock_unlock(&readOnlyGlobals.redis.lock_get);

    if(reply) {
      if(reply->str) {
	ret = strdup(reply->str);
	traceEvent(TRACE_NORMAL, "[Redis] %s(%u)=%s", __FUNCTION__, key, ret);
      }
      freeReplyObject(reply);
    }
  }

  return(ret);
#else
  return(NULL);
#endif
}

/* ************************************ */

#ifdef HAVE_REDIS

static void processQueuedRedisCommand(void) {
  if(readOnlyGlobals.redis.queuedSetDeleteCommands > 0) {
    redisReply *reply;

    pthread_rwlock_wrlock(&readOnlyGlobals.redis.lock_set_delete);
    redisGetReply(readOnlyGlobals.redis.set_delete_context, (void**)&reply);
    freeReplyObject(reply);
    readOnlyGlobals.redis.queuedSetDeleteCommands--;

#if 0
    if(unlikely(readOnlyGlobals.enable_debug))
      traceEvent(TRACE_NORMAL, "[Redis] %s(): got reply [queue=%d]",
		 __FUNCTION__, readOnlyGlobals.redis.queuedSetDeleteCommands);
#endif

    pthread_rwlock_unlock(&readOnlyGlobals.redis.lock_set_delete);
  }
}
#endif

/* ************************************ */

#define MAX_NUM_ARGUMENTS   32
#define MAX_ARGUMENT_LEN   256

int readRedisLine(int sock, char *line) {
  int pos = 0, rc;

  while((rc = recv(sock, &line[pos++], 1, 0)) > 0) {
    if((pos > 2) 
       && (line[pos-2] == '\r')
       && (line[pos-1] == '\n')) {
      line[pos++] = '\0';
      return(pos);
    }
  }

  return(0);
}

int readRedisCommand(int sock, char request[MAX_NUM_ARGUMENTS][MAX_ARGUMENT_LEN]) {
  /*
   *1
   $4
   PING
   
   *<number of arguments> CR LF
   $<number of bytes of argument 1> CR LF
   <argument data> CR LF
   ...
   $<number of bytes of argument N> CR LF
   <argument data> CR LF
  */
  
  
  readRedisLine(sock, request[0]);
  request[1][0] = '\0';

  return(0);
}

/* ************************************ */

void handleClient(int sock) {
  char request[MAX_NUM_ARGUMENTS][MAX_ARGUMENT_LEN], *rsp;
  
  while(readRedisCommand(sock, request) >= 0) {
    int i;

    for(i=0; request[i][0] != '\0'; i++)
      traceEvent(TRACE_NORMAL, "%s", request[i]);

    rsp = "+OK\r\n";
    send(sock, rsp, strlen(rsp), 0);
  }
}

/* ************************************ */

#ifdef HAVE_REDIS
static void* redisLocalServerLoop(void* notUsed) {
  traceEvent(TRACE_NORMAL, "[Redis Server] %s() started", __FUNCTION__);

  readOnlyGlobals.redis.local_server_running = 0;

  while(!readWriteGlobals->shutdownInProgress) {
    fd_set mask;

    FD_ZERO(&mask);
    FD_SET(readOnlyGlobals.redis.local_server_socket, &mask);

    if(select(readOnlyGlobals.redis.local_server_socket+1, &mask, 0, 0, NULL) > 0) {
      struct sockaddr_in from;
      int rx_sock, from_len, rc;
      char aChar;

      rx_sock = accept(readOnlyGlobals.redis.local_server_socket, (struct sockaddr*)&from, (socklen_t*)&from_len);
      if((rx_sock < 0) || (errno != 0)) {
	traceEvent(TRACE_ERROR, "Unable to accept connection [%s/%d]", strerror(errno), errno);	
      }

      traceEvent(TRACE_NORMAL, "[Redis Server] New client connected [socket %d]", rx_sock);
      handleClient(rx_sock);
      close(rx_sock);
    }
  }

  readOnlyGlobals.redis.local_server_running = 1;
}
#endif

/* ************************************ */

#ifdef HAVE_REDIS
static void* redisAsyncLoop(void* notUsed) {
  traceEvent(TRACE_NORMAL, "[Redis] %s() started", __FUNCTION__);

  readOnlyGlobals.redis.queue_thread_running = 1;

  while(!readWriteGlobals->shutdownInProgress) {
    if(readOnlyGlobals.redis.queuedSetDeleteCommands > 0) {
      processQueuedRedisCommand();
    } else {
      usleep(10000); /* 10 ms */
    }
  }

  /* Flush pending commands */
  while(readOnlyGlobals.redis.queuedSetDeleteCommands > 0)
      processQueuedRedisCommand();

  readOnlyGlobals.redis.queue_thread_running = 0;

  traceEvent(TRACE_INFO, "[Redis] %s() completed [queue=%d]",
	     __FUNCTION__, readOnlyGlobals.redis.queuedSetDeleteCommands);
  return(NULL);
}
#endif

/* ************************************ */

int createLocalCacheServer() {
#ifdef HAVE_REDIS
  int sockopt = 1, rc;
  struct sockaddr_in sockIn;
  struct addrinfo hints, *ai = NULL;

  if(readOnlyGlobals.redis.local_ucloud_port == 0)
    return(0);

  errno = 0;
  readOnlyGlobals.redis.local_server_socket = socket(AF_INET, SOCK_STREAM, IPPROTO_TCP);
  
  if((readOnlyGlobals.redis.local_server_socket < 0) || (errno != 0)) {
    traceEvent(TRACE_ERROR, "Unable to create server socket [%s/%d]", strerror(errno), errno);
    exit(-1);
  }

  errno = 0;
  rc = setsockopt(readOnlyGlobals.redis.local_server_socket, SOL_SOCKET, SO_REUSEADDR, (char *)&sockopt, sizeof(sockopt));

  memset(&sockIn, 0, sizeof(sockIn));
  sockIn.sin_family = AF_INET, sockIn.sin_port = (int)htons(readOnlyGlobals.redis.local_ucloud_port), errno = 0;
  rc = bind(readOnlyGlobals.redis.local_server_socket, (struct sockaddr *)&sockIn, sizeof(sockIn));
  if((readOnlyGlobals.redis.local_server_socket < 0) || (errno != 0)) {
    traceEvent(TRACE_ERROR, "Unable to bind to the specified port [%s/%d]", strerror(errno), errno);
    exit(-1);
  }

  errno = 0;
  rc = listen(readOnlyGlobals.redis.local_server_socket, 1 /* 1 client max */);

  pthread_create(&readOnlyGlobals.redis.local_server_loop, NULL, redisLocalServerLoop, NULL);
#endif
  return(0);
}

/* ************************************ */

int connectToRemoteCache(void) {
#ifdef HAVE_REDIS
  struct timeval timeout = { 1, 500000 }; // 1.5 seconds

  if(unlikely(readOnlyGlobals.enable_debug))
    traceEvent(TRACE_NORMAL, "[Redis] %s(%s:%u)", __FUNCTION__, 
	       readOnlyGlobals.redis.remote_redis_host, readOnlyGlobals.redis.remote_redis_port);

  /* Synchronous */
  readOnlyGlobals.redis.context = redisConnectWithTimeout(readOnlyGlobals.redis.remote_redis_host, 
							  readOnlyGlobals.redis.remote_redis_port, timeout);
  if(readOnlyGlobals.redis.context->err) {
    traceEvent(TRACE_ERROR, "Redis Connection error: %s",
	       readOnlyGlobals.redis.context->errstr);
    exit(-1);
  }

  /* Asynchronous */
  readOnlyGlobals.redis.set_delete_context = redisConnectWithTimeout(readOnlyGlobals.redis.remote_redis_host, 
								     readOnlyGlobals.redis.remote_redis_port, timeout);
  if(readOnlyGlobals.redis.set_delete_context->err) {
    traceEvent(TRACE_ERROR, "Redis Connection error: %s",
	       readOnlyGlobals.redis.set_delete_context->errstr);
    exit(-1);
  }

  pthread_rwlock_init(&readOnlyGlobals.redis.lock_get, NULL);
  pthread_rwlock_init(&readOnlyGlobals.redis.lock_set_delete, NULL);

  pthread_create(&readOnlyGlobals.redis.reply_loop, NULL, redisAsyncLoop, NULL);

  createLocalCacheServer();
#endif

  return(0);
}

/* ************************************ */

void disconnectFromRemoteCache(void) {
#ifdef HAVE_REDIS
  if(unlikely(readOnlyGlobals.enable_debug))
    traceEvent(TRACE_NORMAL, "[Redis] %s()", __FUNCTION__);

  while(readOnlyGlobals.redis.queuedSetDeleteCommands > 0) {
    if(!readOnlyGlobals.redis.queue_thread_running)
      processQueuedRedisCommand();
    else
      sleep(1);
  }

  if(readOnlyGlobals.redis.context)
    redisFree(readOnlyGlobals.redis.context);

  if(readOnlyGlobals.redis.set_delete_context)
    redisFree(readOnlyGlobals.redis.set_delete_context);

  pthread_rwlock_destroy(&readOnlyGlobals.redis.lock_get);
  pthread_rwlock_destroy(&readOnlyGlobals.redis.lock_set_delete);

  if(readOnlyGlobals.redis.local_server_socket > 0)
    close(readOnlyGlobals.redis.local_server_socket);
#endif
}

/* ************************************ */

void dumpCacheStats() {
#ifdef HAVE_REDIS
  traceEvent(TRACE_NORMAL, "Cache [queuedSetDelete: actual %u/max %u]",
	     readOnlyGlobals.redis.queuedSetDeleteCommands,
	     readOnlyGlobals.redis.maxQueuedSetDeleteCommands);
#endif
}

/* ************************************ */

void init_cache(struct LruCache *cache, u_int32_t max_size) {
  cache->max_size = max_size, cache->root = NULL;
}

/* ************************************ */

u_int32_t find_in_cache(struct LruCache *cache, u_int64_t key) {
  struct LruCacheEntry *entry;

  HASH_FIND_UINT32_T(cache->root, &key, entry);

  if(entry) {
    /* remove it (so the subsequent add will throw it on the front of the list) */
    HASH_DELETE(hh, cache->root, entry);
    HASH_ADD_KEYPTR(hh, cache->root, &key, sizeof(entry->key), entry);

    if(unlikely(readOnlyGlobals.enable_debug))
      traceEvent(TRACE_NORMAL, "[LRU] %s(%llu)=%u", __FUNCTION__, key, entry->value);
    return(entry->value);
  }

  if(unlikely(readOnlyGlobals.enable_debug))
    traceEvent(TRACE_NORMAL, "[LRU] %s(%llu)=???", __FUNCTION__, key);

  return(NDPI_PROTOCOL_UNKNOWN);
}

/* ************************************ */

int add_to_cache(struct LruCache *cache, u_int64_t key, u_int32_t value)
{
  struct LruCacheEntry *entry, *tmp_entry;

  if(unlikely(readOnlyGlobals.enable_debug))
    traceEvent(TRACE_NORMAL, "[LRU] %s(%llu)=%u", __FUNCTION__, key, value);

  if((entry = malloc(sizeof(struct LruCacheEntry))) == NULL)
    return(-1);

  entry->key = key, entry->value = value;
  HASH_ADD_UINT32_T(cache->root, key, entry);
    
  if (HASH_COUNT(cache->root) >= cache->max_size) {
    HASH_ITER(hh, cache->root, entry, tmp_entry) {
      /* Prune the first (i.e. oldest) entry */
      HASH_DELETE(hh, cache->root, entry);
      free(entry);
      break;
    }
  }    

  return(0);
}

